#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/time.h>

#ifdef TEST_64BIT
typedef unsigned long long uint;
#else
typedef unsigned int uint;
#endif

#define ITERATIONS (1 << 21)
#define SEED {2147483563, 2147483563 ^ 0x49616E42}
#define NANOSEC_PER_SEC 1000000000ULL
#define NANOSEC_PER_USEC 1000ULL

struct random_state {
    uint32_t hi;
    uint32_t lo;
};


static uint64_t nanoseconds(void) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return now.tv_sec * NANOSEC_PER_SEC + now.tv_usec * NANOSEC_PER_USEC;
}

static uint32_t my_random(struct random_state *state) {
    state->hi = (state->hi << 16) + (state->hi >> 16);
    state->hi += state->lo;
    state->lo += state->hi;
    return state->hi;
}

static const uint *random_numerators(void) {
    uint32_t i, *data = (uint32_t *)malloc(ITERATIONS * sizeof(uint));
    struct random_state state = SEED;
    for (i=0; i < ITERATIONS * sizeof(uint) / sizeof(uint32_t); i++) {
        data[i] = my_random(&state);
    }
    return (uint *)data;
}

#define DIVISOR_COUNT 10

#define DIVIDER(D) D,
const uint kDivisors[DIVISOR_COUNT] = {
#include "divider_list.c"
};
#undef DIVIDER


#define DIVIDER(D) uint divide_##D (const uint *nums, size_t count);
#include "divider_list.c"
#undef DIVIDER


#define DIVIDER(D) divide_##D,
uint (* const kFunctions[DIVISOR_COUNT])(const uint *, size_t) = {
#include "divider_list.c"
};
#undef DIVIDER


int main(void) {
    uint results[DIVISOR_COUNT] = {0};
    uint64_t start, times[DIVISOR_COUNT] = {0};
    const void *data = random_numerators();
    const size_t repeatCount = 128;
    size_t i, j;
    
    /* Prewarm */
    for (j=0; j < DIVISOR_COUNT; j++) {
        results[j] += kFunctions[j](data, ITERATIONS);
    }

    /* Test */
    for (i=0; i < repeatCount; i++) {
        for (j=0; j < DIVISOR_COUNT; j++) {
            start = nanoseconds();
            results[j] += kFunctions[j](data, ITERATIONS);
            times[j] += nanoseconds() - start;
        }
    }
    
    /* Output */
    for (j=0; j < DIVISOR_COUNT; j++) {
        printf("%12llu: %.5f nsec (checksum: %llu)\n", (unsigned long long)kDivisors[j], (times[j] / (double)ITERATIONS) / (double)repeatCount, (unsigned long long)results[j]);
    }
    return 0;
}
